import uuid
from datetime import datetime, timezone

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import (
    FeatureUsage,
    GitRefType,
    Process,
    ProcessMode,
    Repository,
    WorkerType,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model
from arkindex.users.models import Role, Scope


class TestDockerWorkerVersion(FixtureAPITestCase):
    """
    Test endpoint to create a worker version from a docker image
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = Repository.objects.create(url="http://my_repo.fake/workers/worker")
        cls.rev = cls.repo.revisions.create(
            hash=str(uuid.uuid4()),
            message="A revision",
            author="Someone",
        )
        cls.repo.workers.create(
            slug="test_worker",
            type=WorkerType.objects.first(),
            description="it's a-me a workerio",
            name="Workerio"
        )
        cls.worker = cls.repo.workers.filter(slug="test_worker").select_related("type").get()
        cls.worker.versions.create(
            revision=cls.rev,
            configuration={},
            docker_image_iid="registry.fake.com/workers/worker:02564",
            state=WorkerVersionState.Available
        )
        cls.version = cls.worker.versions.select_related("revision").get(revision_id=cls.rev.id)
        cls.user.user_scopes.create(scope=Scope.CreateDockerWorkerVersion)
        cls.worker_type = WorkerType.objects.create(slug="docker", display_name="Docker")

    def test_create_requires_login(self):
        response = self.client.post(reverse("api:version-from-docker"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_create_deny_ponos_auth(self):
        """
        Ponos task authentication is disabled on that endpoint
        """
        process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            farm=Farm.objects.first(),
            corpus=self.corpus,
        )
        process.run()
        task = process.tasks.get()
        response = self.client.post(
            reverse("api:version-from-docker"),
            HTTP_AUTHORIZATION=f"Ponos {task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:version-from-docker"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_create_requires_scope(self):
        """
        A specific user scope is required to create a worker version from a docker image
        """
        self.user.user_scopes.get(scope=Scope.CreateDockerWorkerVersion).delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:version-from-docker"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {
            "detail": "You do not have permission to perform this action.",
            "missing_scopes": [Scope.CreateDockerWorkerVersion.value],
        })

    def test_create_admin_no_scope(self):
        """
        User scope is not required for instance admins
        """
        self.assertEqual(self.superuser.user_scopes.count(), 0)
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse("api:version-from-docker"),
            data={
                "configuration": {"test": "A"},
                "docker_image_iid": "a_docker_image",
                "repository_url": self.repo.url,
                "revision_hash": "new_revision_hash",
                "worker_slug": self.worker.slug,
                "worker_name": self.worker.name,
                "worker_type": self.worker.type.slug,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:version-from-docker"))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "docker_image_iid": ["This field is required."],
                "repository_url": ["This field is required."],
                "revision_hash": ["This field is required."],
                "worker_slug": ["This field is required."],
                "worker_name": ["This field is required."],
                "worker_type": ["This field is required."],
            },
        )

    def test_create_fields_validation(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": "str",
                    "docker_image_iid": [],
                    "repository_url": [],
                    "revision_hash": "",
                    "worker_slug": [],
                    "worker_name": [],
                    "worker_type": [],
                    "worker_description": [],
                    "revision_message": [],
                    "revision_author": [],
                    "revision_references": [{"a": 133}],
                    "gpu_usage": "AAAAAAAA",
                    "model_usage": [],
                    # Only revision can be set using this endpoint
                    "version": 42,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "configuration": ['Expected a dictionary of items but got type "str".'],
            "docker_image_iid": ["Not a valid string."],
            "gpu_usage": ["Value is not of type FeatureUsage"],
            "model_usage": ["Value is not of type FeatureUsage"],
            "repository_url": ["Not a valid string."],
            "revision_author": ["Not a valid string."],
            "revision_hash": ["This field may not be blank."],
            "revision_message": ["Not a valid string."],
            "revision_references": [{
                "name": ["This field is required."],
                "type": ["This field is required."]
            }],
            "worker_slug": ["Not a valid string."],
            "worker_name": ["Not a valid string."],
            "worker_type": ["Not a valid string."],
            "worker_description": ["Not a valid string."],
        })

    def test_create_duplicated(self):
        """
        No worker version can be created with an existing revision hash and worker slug
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {},
                    "docker_image_iid": "some_docker_image",
                    "repository_url": self.repo.url,
                    "revision_hash": self.version.revision.hash,
                    "worker_slug": self.worker.slug,
                    "worker_name": self.worker.name,
                    "worker_type": self.worker.type.slug,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": ["A worker version already exists on this revision and this worker slug."]
        })

    def test_create_archived(self):
        self.client.force_login(self.user)
        self.worker.archived = datetime.now(timezone.utc)
        self.worker.save()

        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {"test": "A"},
                    "docker_image_iid": "a_docker_image",
                    "repository_url": self.repo.url,
                    "revision_hash": "new_revision_hash",
                    "worker_slug": self.worker.slug,
                    "worker_name": self.worker.name,
                    "worker_type": self.worker.type.slug,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_slug": ["This worker is archived."],
        })

    def test_create_default_message(self):
        """
        A new version can be published with only the required fields
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(18):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {"test": "A"},
                    "docker_image_iid": "a_docker_image",
                    "repository_url": self.repo.url,
                    "revision_hash": "new_revision_hash",
                    "worker_slug": self.worker.slug,
                    "worker_name": self.worker.name,
                    "worker_type": self.worker.type.slug,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_revision = self.repo.revisions.get(hash="new_revision_hash")
        new_version = new_revision.versions.get()
        self.assertEqual(new_version.worker, self.worker)
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "A"},
            "docker_image_iid": "a_docker_image",
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Disabled.value,
            "revision": {
                "id": str(new_revision.id),
                "author": "default",
                "commit_url": "http://my_repo.fake/workers/worker/commit/new_revision_hash",
                "created": new_revision.created.isoformat().replace("+00:00", "Z"),
                "hash": "new_revision_hash",
                "message": "created from docker image",
                "refs": []
            },
            "revision_url": None,
            "version": None,
            "tag": None,
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "state": "available",
            "worker": {
                "id": str(self.worker.id),
                "name": self.worker.name,
                "slug": self.worker.slug,
                "type": self.worker.type.slug
            }
        })

    def test_create(self):
        """
        A new version can be published with the optional values
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(19):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {"test": "A"},
                    "docker_image_iid": "e" * 512,
                    "repository_url": self.repo.url,
                    "revision_hash": "new_revision_hash",
                    "worker_slug": self.worker.slug,
                    "worker_name": self.worker.name,
                    "worker_type": self.worker.type.slug,
                    "revision_message": "Bruce was very clever",
                    "revision_author": "Iwan Roberts",
                    "revision_references": [
                        {"type": "branch", "name": "master"},
                        {"type": "tag", "name": "2.0"},
                    ],
                    "gpu_usage": FeatureUsage.Required.value,
                    "model_usage": FeatureUsage.Supported.value,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_revision = self.repo.revisions.get(hash="new_revision_hash")
        refs = list(new_revision.refs.all())
        self.assertDictEqual(
            {ref.type: ref.name for ref in refs},
            {GitRefType.Branch: "master", GitRefType.Tag: "2.0"},
        )
        new_version = new_revision.versions.get()
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "A"},
            "docker_image_iid": "e" * 512,
            "gpu_usage": FeatureUsage.Required.value,
            "model_usage": FeatureUsage.Supported.value,
            "revision": {
                "id": str(new_revision.id),
                "author": "Iwan Roberts",
                "commit_url": "http://my_repo.fake/workers/worker/commit/new_revision_hash",
                "created": new_revision.created.isoformat().replace("+00:00", "Z"),
                "hash": "new_revision_hash",
                "message": "Bruce was very clever",
                "refs": [
                    {"id": str(ref.id), "name": ref.name, "type": ref.type.value}
                    for ref in refs
                ],
            },
            "revision_url": None,
            "version": None,
            "tag": None,
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "state": "available",
            "worker": {
                "id": str(self.worker.id),
                "name": self.worker.name,
                "slug": self.worker.slug,
                "type": self.worker.type.slug
            }
        })
        # Existing repository memberships are not updated
        self.assertFalse(self.repo.memberships.exists())

    def test_create_update_worker(self):
        """
        Creating a new worker version can update the worker, and create a new worker type
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(23):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {"test": "A"},
                    "docker_image_iid": "e" * 512,
                    "repository_url": self.repo.url,
                    "revision_hash": "new_revision_hash",
                    "worker_slug": self.worker.slug,
                    "worker_name": "A New Name",
                    "worker_type": "new_worker_type",
                    "worker_description": "C'est un petit val qui mousse de rayons.",
                    "revision_message": "Bruce was very clever",
                    "revision_author": "Iwan Roberts",
                    "revision_references": [
                        {"type": "branch", "name": "master"},
                        {"type": "tag", "name": "2.0"},
                    ],
                    "gpu_usage": FeatureUsage.Required.value,
                    "model_usage": FeatureUsage.Supported.value,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_worker_type = WorkerType.objects.get(slug="new_worker_type")
        new_revision = self.repo.revisions.get(hash="new_revision_hash")
        self.worker.refresh_from_db()
        self.assertEqual(self.worker.description, "C'est un petit val qui mousse de rayons.")
        refs = list(new_revision.refs.all())
        self.assertDictEqual(
            {ref.type: ref.name for ref in refs},
            {GitRefType.Branch: "master", GitRefType.Tag: "2.0"},
        )
        new_version = new_revision.versions.get()
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "A"},
            "docker_image_iid": "e" * 512,
            "gpu_usage": FeatureUsage.Required.value,
            "model_usage": FeatureUsage.Supported.value,
            "revision": {
                "id": str(new_revision.id),
                "author": "Iwan Roberts",
                "commit_url": "http://my_repo.fake/workers/worker/commit/new_revision_hash",
                "created": new_revision.created.isoformat().replace("+00:00", "Z"),
                "hash": "new_revision_hash",
                "message": "Bruce was very clever",
                "refs": [
                    {"id": str(ref.id), "name": ref.name, "type": ref.type.value}
                    for ref in refs
                ],
            },
            "revision_url": None,
            "version": None,
            "tag": None,
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "state": "available",
            "worker": {
                "id": str(self.worker.id),
                "name": "A New Name",
                "slug": self.worker.slug,
                "type": new_worker_type.slug
            }
        })
        # Existing repository memberships are not updated
        self.assertFalse(self.repo.memberships.exists())

    def test_create_update_git_ref(self):
        """
        Existing GitRefs on the repository on a different revision
        are updated to the new revision
        """
        self.assertFalse(self.repo.refs.exists())
        branch_ref = self.repo.refs.create(
            revision=self.rev,
            type=GitRefType.Branch,
            name="master",
        )
        tag_ref = self.repo.refs.create(
            revision=self.rev,
            type=GitRefType.Tag,
            name="2.0",
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(19):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {"test": "A"},
                    "docker_image_iid": "a_docker_image",
                    "repository_url": self.repo.url,
                    "revision_hash": "new_revision_hash",
                    "worker_slug": self.worker.slug,
                    "worker_name": self.worker.name,
                    "worker_type": self.worker.type.slug,
                    "revision_message": "Bruce was very clever",
                    "revision_author": "Iwan Roberts",
                    "revision_references": [
                        {"type": "branch", "name": "master"},
                        {"type": "tag", "name": "2.0"},
                    ],
                    "gpu_usage": FeatureUsage.Required.value,
                    "model_usage": FeatureUsage.Required.value,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_revision = self.repo.revisions.get(hash="new_revision_hash")

        branch_ref.refresh_from_db()
        tag_ref.refresh_from_db()
        self.assertEqual(branch_ref.revision, new_revision)
        self.assertEqual(tag_ref.revision, new_revision)

        # No new refs were created
        self.assertEqual(self.repo.refs.count(), 2)

        data = response.json()
        # GitRefs are not sorted in the API output, so we have to assert separately
        refs_data = data["revision"].pop("refs")
        self.assertCountEqual(refs_data, [
            {
                "id": str(tag_ref.id),
                "name": "2.0",
                "type": "tag",
            },
            {
                "id": str(branch_ref.id),
                "name": "master",
                "type": "branch",
            },
        ])

        new_version = new_revision.versions.get()
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "A"},
            "docker_image_iid": "a_docker_image",
            "gpu_usage": FeatureUsage.Required.value,
            "model_usage": FeatureUsage.Required.value,
            "revision": {
                "id": str(new_revision.id),
                "author": "Iwan Roberts",
                "commit_url": "http://my_repo.fake/workers/worker/commit/new_revision_hash",
                "created": new_revision.created.isoformat().replace("+00:00", "Z"),
                "hash": "new_revision_hash",
                "message": "Bruce was very clever",
            },
            "revision_url": None,
            "version": None,
            "tag": None,
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "state": "available",
            "worker": {
                "id": str(self.worker.id),
                "name": self.worker.name,
                "slug": self.worker.slug,
                "type": self.worker.type.slug
            }
        })
        # Existing repository memberships are not updated
        self.assertFalse(self.repo.memberships.exists())

    def test_create_add_git_ref(self):
        """
        If a worker version already exists, the call still succeeds
        when it creates a new GitRef
        """
        self.client.force_login(self.user)
        self.assertFalse(self.version.revision.refs.exists())
        branch_ref = self.version.revision.refs.create(
            name="fix-typo",
            type=GitRefType.Branch,
            repository=self.version.revision.repo,
        )

        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {},
                    "docker_image_iid": "some_docker_image",
                    "repository_url": self.repo.url,
                    "revision_hash": self.version.revision.hash,
                    "worker_slug": self.worker.slug,
                    "worker_name": self.worker.name,
                    "worker_type": self.worker.type.slug,
                    "revision_references": [
                        {"type": "tag", "name": "9.9.9.9.9.9"},
                    ],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        tag_ref = self.version.revision.refs.get(type=GitRefType.Tag)

        self.assertDictEqual(response.json(), {
            "id": str(self.version.id),
            "configuration": {},
            "docker_image_iid": self.version.docker_image_iid,
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Disabled.value,
            "revision": {
                "id": str(self.version.revision.id),
                "author": "Someone",
                "commit_url": f"http://my_repo.fake/workers/worker/commit/{self.version.revision.hash}",
                "created": self.version.revision.created.isoformat().replace("+00:00", "Z"),
                "hash": self.version.revision.hash,
                "message": "A revision",
                "refs": [
                    {
                        "id": str(branch_ref.id),
                        "name": "fix-typo",
                        "type": "branch",
                    },
                    {
                        "id": str(tag_ref.id),
                        "name": "9.9.9.9.9.9",
                        "type": "tag",
                    },
                ],
            },
            "revision_url": None,
            "version": None,
            "tag": None,
            "created": self.version.created.isoformat().replace("+00:00", "Z"),
            "state": "available",
            "worker": {
                "id": str(self.worker.id),
                "name": self.worker.name,
                "slug": self.worker.slug,
                "type": self.worker.type.slug
            }
        })

    def test_create_git_stack(self):
        """
        All git references can be created including repository, revision, gitrefs and worker
        and worker type
        """
        self.worker_type.delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(32):
            response = self.client.post(
                reverse("api:version-from-docker"),
                data={
                    "configuration": {"key": "value"},
                    "docker_image_iid": "docker_image_42",
                    "repository_url": "https://gitlab.test.arkindex.org/project/",
                    "revision_hash": "deadbeef",
                    "worker_slug": "new_gen_classifier",
                    "worker_name": "Classifier New Generation",
                    "worker_type": "a_new_type",
                    "revision_references": [
                        {"type": "branch", "name": "master"},
                    ],
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_repo = Repository.objects.get(url="https://gitlab.test.arkindex.org/project/")
        new_revision = new_repo.revisions.get(hash="deadbeef")
        new_ref = new_revision.refs.get()
        new_version = new_revision.versions.get()
        new_worker_type = new_version.worker.type
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"key": "value"},
            "docker_image_iid": "docker_image_42",
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Disabled.value,
            "revision": {
                "id": str(new_revision.id),
                "author": "default",
                "commit_url": "https://gitlab.test.arkindex.org/project/commit/deadbeef",
                "created": new_revision.created.isoformat().replace("+00:00", "Z"),
                "hash": "deadbeef",
                "message": "created from docker image",
                "refs": [{"id": str(new_ref.id), "name": new_ref.name, "type": new_ref.type.value}],
            },
            "state": "available",
            "revision_url": None,
            "version": None,
            "tag": None,
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "worker": {
                "id": str(new_version.worker.id),
                "name": "Classifier New Generation",
                "slug": "new_gen_classifier",
                "type": str(new_worker_type),
            }
        })
        # Check attributes directly from the DB
        self.assertEqual(new_repo.url, "https://gitlab.test.arkindex.org/project/")
        self.assertEqual(new_revision.author, "default")
        self.assertEqual(new_revision.message, "created from docker image")
        self.assertEqual(new_revision.repo, new_repo)
        self.assertEqual(new_ref.name, "master")
        self.assertEqual(new_ref.repository, new_repo)
        self.assertEqual(new_version.worker.name, "Classifier New Generation")
        self.assertEqual(new_version.worker.slug, "new_gen_classifier")
        self.assertEqual(new_version.worker.public, False)
        self.assertEqual(new_version.worker.repository, new_repo)
        self.assertEqual(new_version.configuration, {"key": "value"})
        self.assertEqual(new_version.docker_image_iid, "docker_image_42")
        self.assertEqual(new_version.gpu_usage, FeatureUsage.Disabled)
        self.assertEqual(new_version.model_usage, FeatureUsage.Disabled)
        self.assertEqual(new_worker_type.slug, "a_new_type")
        self.assertEqual(new_worker_type.display_name, "a_new_type")
        # User is granted an admin role on the repository
        self.assertListEqual(list(new_repo.memberships.values_list("user", "level")), [
            (self.user.id, Role.Admin.value)
        ])

    # Test user configuration

    def test_create_version_valid_user_configuration(self):
        test_model = Model.objects.create(
            name="Generic model",
            public=False,
        )
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:version-from-docker"),
            data={
                "docker_image_iid": "a_docker_image",
                "repository_url": self.repo.url,
                "revision_hash": "new_revision_hash",
                "worker_slug": self.worker.slug,
                "worker_name": self.worker.name,
                "worker_type": self.worker.type.slug,
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"title": "Demo Integer", "type": "int", "required": True, "default": 1},
                        "demo_boolean": {"title": "Demo Boolean", "type": "bool", "required": False, "default": True},
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": "b", "c": "d"}},
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": [1, 2, 3]},
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, 3, 4]},
                        "boolean_list": {"title": "It's a list of booleans", "type": "list", "required": False, "subtype": "bool", "default": [True, False, False]},
                        "demo_model": {"title": "Model for training", "type": "model", "required": True},
                        "other_model": {"title": "Model the second", "type": "model", "default": str(test_model.id)}
                    }
                },
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_integer": {
                    "title": "Demo Integer",
                    "type": "int",
                    "required": True,
                    "default": 1
                },
                "demo_boolean": {
                    "title": "Demo Boolean",
                    "type": "bool",
                    "required": False,
                    "default": True
                },
                "demo_dict": {
                    "title": "Demo Dict",
                    "type": "dict",
                    "required": True,
                    "default": {"a": "b", "c": "d"}
                },
                "demo_choice": {
                    "title": "Decisions",
                    "type": "enum",
                    "required": True,
                    "default": 1,
                    "choices": [1, 2, 3]
                },
                "demo_list": {
                    "title": "Demo List",
                    "type": "list",
                    "subtype": "int",
                    "required": True,
                    "default": [1, 2, 3, 4]
                },
                "boolean_list": {
                    "title": "It's a list of booleans",
                    "type": "list",
                    "subtype": "bool",
                    "required": False,
                    "default": [True, False, False]
                },
                "demo_model": {
                    "title": "Model for training",
                    "type": "model",
                    "required": True
                },
                "other_model": {
                    "title": "Model the second",
                    "type": "model",
                    "default": str(test_model.id)
                }
            }
        })

    def test_create_invalid_user_configuration(self):
        cases = [
            (
                "non",
                ['Expected a dictionary of items but got type "str".']
            ),
            (
                {"demo_list": {"title": "Demo List", "type": "list", "required": True, "default": [1, 2, 3, 4]}},
                {"demo_list": {
                    "subtype": ['The "subtype" field must be set for "list" type properties.']
                }}
            ),
            (
                {"demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "dict", "default": [1, 2, 3, 4]}},
                {"demo_list": {
                    "subtype": ["Subtype can only be int, float, bool or string."]
                }}
            ),
            (
                {"demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, "three", 4]}},
                {"demo_list": {
                    "default": ["All items in the default value must be of type int."]
                }}
            ),
            (
                {"demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": "eeee"}},
                {"demo_choice": {
                    "choices": ['Expected a list of items but got type "str".']
                }}
            ),
            (
                {"secrets": ["aaaaaa"]},
                {"secrets": {"__all__": ["User configuration field definitions should be of type dict, not list."]}}
            ),
            (
                {"something": {"title": "some thing", "type": "uh oh", "required": 2}},
                {"something": {
                    "required": ["Must be a valid boolean."],
                    "type": ["Value is not of type UserConfigurationFieldType"]
                }}
            ),
            (
                {"something": {"title": "some thing", "type": "int", "required": False, "choices": [1, 2, 3]}},
                {"something": {
                    "choices": ['The "choices" field can only be set for an "enum" type property.']
                }}
            ),
            (
                {"demo_integer": {"type": "int", "required": True, "default": 1}},
                {"demo_integer": {"title": ["This field is required."]}}
            ),
            (
                {"demo_integer": {"title": "an integer", "type": "int", "required": True, "default": 1, "some_key": "oh no"}},
                {"demo_integer": {
                    "some_key": ["Configurable properties can only be defined using the following keys: title, type, required, default, subtype, choices."]
                }}
            ),
            (
                {"param": {"title": "Model to train", "type": "model", "default": "12341234-1234-1234-1234-123412341234"}},
                {"param": {"default": ["Model 12341234-1234-1234-1234-123412341234 not found."]}}
            )
        ]

        self.client.force_login(self.user)
        for user_configuration, error in cases:
            with self.subTest(user_configuration=user_configuration, error=error):
                response = self.client.post(
                    reverse("api:version-from-docker"),
                    data={
                        "docker_image_iid": "a_docker_image",
                        "repository_url": self.repo.url,
                        "revision_hash": "new_revision_hash",
                        "worker_slug": self.worker.slug,
                        "worker_name": self.worker.name,
                        "worker_type": self.worker.type.slug,
                        "configuration": {
                            "user_configuration": user_configuration
                        },
                    },
                    format="json",
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {
                    "configuration": {
                        "user_configuration": [error]
                    }
                })

    def test_create_version_invalid_user_configuration_default_value(self):
        self.client.force_login(self.user)
        cases = [
            ({"type": "int", "default": False}, "int"),
            ({"type": "int", "default": True}, "int"),
            ({"type": "float", "default": False}, "float"),
            ({"type": "float", "default": True}, "float"),
            ({"type": "bool", "default": 0}, "bool"),
            ({"type": "bool", "default": 1}, "bool"),
            ({"type": "string", "default": 1}, "string"),
            ({"type": "dict", "default": ["a", "b"]}, "dict"),
            ({"type": "model", "default": "gigi hadid"}, "model"),
            ({"type": "model", "default": False}, "model"),
            ({"type": "list", "subtype": "int", "default": 12}, "list")
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.post(
                    reverse("api:version-from-docker"),
                    data={
                        "docker_image_iid": "a_docker_image",
                        "repository_url": self.repo.url,
                        "revision_hash": "new_revision_hash",
                        "worker_slug": self.worker.slug,
                        "worker_name": self.worker.name,
                        "worker_type": self.worker.type.slug,
                        "configuration": {
                            "user_configuration": {
                                "param": {"title": "param", **params}
                            }
                        },
                    },
                    format="json",
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {"configuration": {"user_configuration": [{"param": {"default": [f"This is not a valid value for a field of type {expected}."]}}]}})
